//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA file format driver in C++      ~
// ~             built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_ilda.hpp"
#include "ez_wave.hpp"
    
//############################################################################
ez_ilda_frame_set::ez_ilda_frame_set(ez_ilda_point(*F)(int, int), int points_per_frame, int num_frames)
{
    for(frame_index = 0; frame_index < num_frames; frame_index++)
        push_back(ez_ilda_frame(F, points_per_frame, frame_index));
    renumber();
    frame_index = 0;
    error = EZ_ILDA_OK;
}

//############################################################################
ez_ilda_frame_set& ez_ilda_frame_set::operator += (const ez_ilda_frame& frame)
{
    if(ez_ilda_pointer->show_effects_generation)
    {
        frame.render_in_ezfb_bitmap();
        bmp_to_SDL_Surface(&ez_ilda_pointer->bmp, screen);
    }
    push_back(frame);
    renumber();
    return *this;
}

//############################################################################
int ez_ilda_frame_set::from_file(string file, bool append)
{
    ez_ilda_pointer->process_exception = false;
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".ild"
      )
        file += ".ild";

    ifstream in((EZ_ILDA_FRAME_SHARE + file).c_str(), ios::in | ios::binary);
    if(in.is_open())
    {
        frame_index = 0;
        file_name   = file;
        error       = EZ_ILDA_OK;
        error       = from_ifstream(in, append);
        in.close();
    }
    else
    {
        ez_ilda_pointer->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "[any key] %s failed to open"
                    , file.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_ifstream(ifstream& in, bool append)
{
    ez_ilda_header  header                     ;
    ez_ilda_frame   frame                      ;
    ez_palette      palette                    ;
    bool            true_color_data    = false ,
                    some_data_accepted = false ;
    //------------------------------------------------------------------------
    if(!append)
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    ez_ilda_pointer->palette_index = EZ_ILDA_DEFAULT;
    //------------------------------------------------------------------------
    while(header.from_ifstream(in) == EZ_ILDA_OK)
    {   
        if(header.quantity)
        {
            switch(header.format)
            {
                //------------------------------------------------------------
                case EZ_ILDA_3D_FRAME_DATA: 
                case EZ_ILDA_2D_FRAME_DATA:
                    error |= frame.from_ifstream(in, header);
                    if(!error)
                    {
                        frame.palette_index = ez_ilda_pointer->palette_index;
                        frame.set_rgb_from_palette();
                        frame.is_2D() ? num_2D_frames++ : num_3D_frames++;
                        push_back(frame);
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
                case EZ_ILDA_PALETTE_DATA:
                    error |= palette.from_ifstream(in, header);
                    if(!error)
                    {
                        bool built_in_palette = false;
                        for(int i = 0; i < EZ_BUILT_IN_PALETTES; i++)
                            if(header.name == ez_ilda_pointer->palette_picker(i).name)
                            {
                                ez_ilda_pointer->palette_index = i;
                                built_in_palette = true;
                                break;
                            }
                        if(!built_in_palette)
                        {
                            ez_ilda_pointer->push_back_palette(palette);
                            ez_ilda_pointer->palette_index = ez_ilda_pointer->number_of_palettes() - 1;
                        }
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
                case EZ_ILDA_TRUE_COLOR_DATA:   
                    error |= frame.true_color_from_ifstream(in, header);
                    if(!error)
                    {
                        true_color_data = true;
                        frame.is_2D() ? num_2D_frames++ : num_3D_frames++;
                        push_back(frame);
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
            }
        }
    }
    //------------------------------------------------------------------------
    if(!in.eof() && error && some_data_accepted)
        error = EZ_ILDA_OK;
    //------------------------------------------------------------------------
    if(true_color_data)
        ez_ilda_pointer->palette_index = EZ_TRUE_COLOR;
    //------------------------------------------------------------------------
    return error;
} 

//############################################################################
int ez_ilda_frame_set::from_wave_file(string file, bool append)
{
    ez_ilda_pointer->process_exception = false;
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".wav"
      )
        file += ".wav";

    ifstream in((EZ_ILDA_WAVE_SHARE + file).c_str(), ios::in | ios::binary);
    if(in.is_open())
    {
        frame_index = 0;
        file_name   = file;
        error       = from_ifstream_wave(in, append);
        if(in.is_open())
            in.close();
    }
    else
    {
        ez_ilda_pointer->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "[any key] %s failed to open"
                    , file.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    return error;
}
//############################################################################
int ez_ilda_frame_set::from_ADAT_wave_files(string file_1,
                                            string file_2,
                                            string file_3,
                                            string file_4
                                           )
{
    ifstream in_1,
             in_2,
             in_3,
             in_4;

    ez_ilda_pointer->process_exception = false;

    if(    file_1.size() < 4
        || file_1.substr(file_1.size() - 4) != ".wav"
      )
        file_1 += ".wav";
    if(    file_2.size() < 4
        || file_2.substr(file_2.size() - 4) != ".wav"
      )
        file_2 += ".wav";
    if(    file_3.size() < 4
        || file_3.substr(file_3.size() - 4) != ".wav"
      )
        file_3 += ".wav";
    if(    file_4.size() < 4
        || file_4.substr(file_4.size() - 4) != ".wav"
      )
        file_4 += ".wav";
        
    in_1.open((EZ_ILDA_ADAT_SHARE + file_1).c_str(), ios::in | ios::binary);
    in_2.open((EZ_ILDA_ADAT_SHARE + file_2).c_str(), ios::in | ios::binary);
    in_3.open((EZ_ILDA_ADAT_SHARE + file_3).c_str(), ios::in | ios::binary);
    in_4.open((EZ_ILDA_ADAT_SHARE + file_4).c_str(), ios::in | ios::binary);
       
    if(    in_1.is_open()
        && in_2.is_open()
        && in_3.is_open()
        && in_4.is_open()
      )
    {
        //error       = measure_ADAT_waves(in_1, in_2, in_3);
        //in_1.seekg(0, ios::beg);
        //in_2.seekg(0, ios::beg);
        //in_3.seekg(0, ios::beg);
 
        frame_index = 0;
        file_name   = file_1;
        error       = from_ifstreams_ADAT_waves(in_1, in_2, in_3, in_4);
        in_1.close();
        in_2.close();
        in_3.close();
        in_4.close();
    }
    else
    {
        ez_ilda_pointer->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "[any key] at least one of those files failed to open"
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_ifstream_wave(ifstream& in, bool append)
{
    ez_wave_header header;
    //------------------------------------------------------------------------
    if((error = header.from_ifstream(in)))
    {
        ez_ilda_pointer->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "[any key] error reading wave header"
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }        
    //------------------------------------------------------------------------
    else if(header.num_channels < 6)
    {
        ez_ilda_pointer->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "[any key] incompatible wave format"
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_ILDA_INCOMPATIBLE_FORMATS;
    }
    //------------------------------------------------------------------------
    else
    {
        int            i         ,
                       span      ,
                       last      ,
                       red       ,
                       green     ,
                       blue      ,
                       offset_max,
                       offset_min;
        ez_ilda_frame  frame     ;
        ez_ilda_point  point     ;
        ez_wave_sample_pointer  *sample_window;
        //--------------------------------------------------------------------
        offset_min = header.offset[0];
        offset_max = header.offset[0];
        for(i = 1; i < 6; i++)
        {
            if(header.offset[i] > offset_max)
                offset_max = header.offset[i];
            if(header.offset[i] < offset_min)
                offset_min = header.offset[i];
        }
        span  = (offset_max - offset_min) + 2;
        last  = span - 1                     ;
        red   = -header.offset[2] + 1        ;
        green = -header.offset[3] + 1        ;
        blue  = -header.offset[4] + 1        ;
        //--------------------------------------------------------------------
        sample_window = new ez_wave_sample_pointer[span];
        for(i = 0; i < span; i++)
            sample_window[i] = new ez_wave_sample(header.num_channels);
        //--------------------------------------------------------------------
        frame.format        = EZ_ILDA_2D_FRAME_DATA;
        frame.palette_index = EZ_TRUE_COLOR;
        //--------------------------------------------------------------------
        if(!append)
        {
            clear();
            num_2D_frames = 0;
            num_3D_frames = 0;
        }
        //--------------------------------------------------------------------
        while(true)
        {
            for(i = 0; i < last; i++)
                sample_window[i] = sample_window[i + 1];
            if(sample_window[last]->from_ifstream(in)) // if EOF
                break;
            point.x = sample_window[1]->channel[0];
            point.y = sample_window[1]->channel[1];
            point.z = 0;
            point.r = 255;
            point.g = 255;
            point.b = 255;
            point.k = 0;
            point.c = 0;
            //----------------------------------------------------------------
            point.r = (u_char)(sample_window[red  ]->channel[2] >> 7);
            point.g = (u_char)(sample_window[green]->channel[3] >> 7);
            point.b = (u_char)(sample_window[blue ]->channel[4] >> 7);
            if(point.r + point.g + point.b)
                point.k = 0;
            else
                point.k = EZ_ILDA_BLANKING_BIT;
            point.c = 0;
//            if(sample_window[blue]->channel[4] & 0x0001)
//                point.set_unique();
            //----------------------------------------------------------------
            frame.push_back(point);
            if(sample_window[red]->channel[2] & 0x0001) // end of frame
            {
                if(sample_window[green]->channel[3] & 0x0001) // frame is unique
                    frame.is_unique = true;
                else
                    frame.is_unique = false;
                frame.quantity = (u_short)frame.number_of_points();
                push_back(frame);
                frame.clear();
            }
        }
        //--------------------------------------------------------------------
        renumber();
        //--------------------------------------------------------------------
//        for(i = 0; i < span; i++)
//            delete sample_window[i];
        delete sample_window;
        error = EZ_ILDA_OK;
        //--------------------------------------------------------------------
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::measure_ADAT_waves(//ifstream& in_1,
                                          ifstream& in_2,
                                          ifstream& in_3,
                                          ifstream& in_4
                                         )
{
    ez_wave_header //header_1,
                   header_2,
                   header_3,
                   header_4;
    ez_wave_sample //sample_1(2),
                   sample_2(1),
                   sample_3(1),
                   sample_4(1);
    //------------------------------------------------------------------------
//    if((error = header_1.from_ifstream(in_1)))
//        return error;
    if((error = header_2.from_ifstream(in_2)))
        return error;
    if((error = header_3.from_ifstream(in_3)))
        return error;
    if((error = header_4.from_ifstream(in_4)))
        return error;
    //------------------------------------------------------------------------
    max_r = EZ_MIN_SHORT;
    max_g = EZ_MIN_SHORT;
    max_b = EZ_MIN_SHORT;
    while(true)
    {
        if(    sample_2.from_ifstream_inverted(in_2)
            || sample_3.from_ifstream_inverted(in_3)
            || sample_4.from_ifstream_inverted(in_4)
          )
            break;                                        

        if(max_r < sample_2.channel[0])
            max_r = sample_2.channel[0];
            
        if(max_g < sample_3.channel[0])
            max_g = sample_3.channel[0];
            
        if(max_b < sample_4.channel[0])
            max_b = sample_4.channel[0];            
    }
    //------------------------------------------------------------------------
    return EZ_ILDA_OK;
}

//############################################################################
int ez_ilda_frame_set::from_ifstreams_ADAT_waves(ifstream& in_1,
                                                 ifstream& in_2,
                                                 ifstream& in_3,
                                                 ifstream& in_4
                                                )
{
    int            i,
                   color_sum,
                   static_sample_count = 0;
                   
    ez_ilda_frame  frame;
    
    ez_ilda_point  point,
                   point_1,
                   point_2,
                   point_3,
                   point_4;
    
    ez_wave_header header_1,
                   header_2,
                   header_3,
                   header_4;
                   
    ez_wave_sample sample_1(2),
                   sample_2(1),
                   sample_3(1),
                   sample_4(1);
                   
//    float          red_gain   = (EZ_MAX_SHORT / max_r) / 128.0,
//                   green_gain = (EZ_MAX_SHORT / max_g) / 128.0,
//                   blue_gain  = (EZ_MAX_SHORT / max_b) / 128.0;

    float          red_gain   = 1.0 / 128.0,
                   green_gain = 1.0 / 128.0,
                   blue_gain  = 1.0 / 128.0;
    //------------------------------------------------------------------------
    frame.format        = EZ_ILDA_2D_FRAME_DATA;
    frame.palette_index = EZ_TRUE_COLOR;
    num_2D_frames       = 0;
    num_3D_frames       = 0;
    clear();
    //------------------------------------------------------------------------
    if((error = header_1.from_ifstream(in_1)))
        return error;
    if((error = header_2.from_ifstream(in_2)))
        return error;
    if((error = header_3.from_ifstream(in_3)))
        return error;
    if((error = header_4.from_ifstream(in_4)))
        return error;
    //------------------------------------------------------------------------
    for(i = 0; i < 7; i++) // color time alignment woo hoo!
    {
        sample_2.from_ifstream_inverted(in_2); // eat-um'up!
        sample_3.from_ifstream_inverted(in_3); // eat-um'up!
        sample_4.from_ifstream_inverted(in_4); // eat-um'up!
    }
    //------------------------------------------------------------------------
    while(true)
    {
        if(    sample_1.from_ifstream_inverted(in_1)
            || sample_2.from_ifstream_inverted(in_2)
            || sample_3.from_ifstream_inverted(in_3)
            || sample_4.from_ifstream_inverted(in_4)
          )
            break;
            
        point.x = sample_1.channel[0];
        point.y = sample_1.channel[1];
        point.z = 0;

        color_sum =   sample_2.channel[0]
                    + sample_3.channel[0]
                    + sample_4.channel[0];

        if(color_sum < 400) // very close to black!
        {
            point.r = 255; // if unblanked it is red
            point.g = 0;
            point.b = 0;
            point.k = EZ_ILDA_BLANKING_BIT; // blank it
            point.c = 0; // 24 bit

            if(    (point | point_2) <= ez_ilda_pointer->insignificant_distance
                && (point | point_4) <= ez_ilda_pointer->insignificant_distance
              )
                static_sample_count++;

            if(static_sample_count > 150) // we found a frame boundary
            {
                frame.is_unique = true;
                frame.quantity = (u_short)frame.number_of_points();
                if(frame.number_of_segments())
                    push_back(frame);
                frame.clear();
                static_sample_count = 0;
                while(true) // get through all of this bunch of blank
                {
                   if(     sample_1.from_ifstream_inverted(in_1)
                        || sample_2.from_ifstream_inverted(in_2)
                        || sample_3.from_ifstream_inverted(in_3)
                        || sample_4.from_ifstream_inverted(in_4)
                      )
                        break;
                    color_sum =   sample_2.channel[0]
                                + sample_3.channel[0]
                                + sample_4.channel[0];

                    if(color_sum > 400) // we have color!
                        break;
                }
            }
            point_4 = point_3;
            point_3 = point_2;
            point_2 = point_1;
            point_1 = point;
        }
        else
        {
            point.r = (u_char)(sample_2.channel[0] * red_gain  );
            point.g = (u_char)(sample_3.channel[0] * green_gain);
            point.b = (u_char)(sample_4.channel[0] * blue_gain );
            point.k = 0;
            point.c = 0;
            static_sample_count = 0;
        }
        //--------------------------------------------------------------------
        frame.push_back(point);
        //--------------------------------------------------------------------
    }
    renumber();
    //------------------------------------------------------------------------
    return EZ_ILDA_OK;
}

//############################################################################
int ez_ilda_frame_set::from_dxf_file(string file, bool append)
{
    ez_ilda_pointer->process_exception = false;
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".dxf"
      )
        file += ".dxf";
    ifstream in((EZ_ILDA_DXF_SHARE + file).c_str(), ios::in);
    if(in.is_open())
    {
        frame_index = 0;
        file_name   = file;
        error       = from_ifstream_dxf(in, append);
        if(error)
        {
            ez_ilda_pointer->process_exception = true;
            bmp_printf(   &BMP
                        , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                        , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                        , BMP.white
                        , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "[any key] failed to find image in %s"
                        , file.c_str()
                      );
            bmp_to_SDL_Surface(&BMP, screen);
        }
        in.close();
    }
    else
    {
        ez_ilda_pointer->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "[any key] %s failed to open"
                    , file.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_dxf_directory(string dir, bool append)
{
    int                     i             ,
                            j             ;
    float                   real_size  = 0,
                            real_scale = 1;
    ez_3D_float             real_min      ,
                            real_max      ,
                            real_offset   ;
    ez_ilda_point           point         ;
    ez_ilda_frame           frame(EZ_DXF) ; // DXF palette
    DIR                    *pDIR          ;
    struct dirent          *pDirEnt       ;
    vector<string>          file_list     ;
    ez_real_segment         real_points   ;
    vector<ez_real_segment> real_segments ;    
    
    pDIR = opendir((EZ_ILDA_DXF_SHARE + dir).c_str());
    ez_ilda_pointer->process_exception = false;
    
    if(pDIR == NULL)
    {
        ez_ilda_pointer->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "[any key] %s failed to open"
                    , dir.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    else
    {
        pDirEnt = readdir(pDIR);  
        while(pDirEnt != NULL)
        {
            if(    strcmp(pDirEnt->d_name, ".")  // not the dot
                && strcmp(pDirEnt->d_name, "..") // not the dot dot
              )
                file_list.push_back(pDirEnt->d_name);
            pDirEnt = readdir(pDIR);
        }
        closedir(pDIR);    
        if(file_list.size())
        {    
            //----------------------------------------------------------------        
            if(!append)
            {
                clear();
                num_2D_frames = 0;
                num_3D_frames = 0;
            }
            //----------------------------------------------------------------
            for(i = 0; i < (int)file_list.size(); i++)
            {
                ifstream in((EZ_ILDA_DXF_SHARE + dir + '/' + file_list[i]).c_str(), ios::in);
                if(in.is_open())
                {
                    real_points.from_ifstream_dxf(in);
                    real_segments.push_back(real_points);
                    real_points.clear();
                    in.close();
                }
            }
            //----------------------------------------------------------------
            if(ez_ilda_pointer->auto_scale_dxf)
            {
                real_min =  FLT_MAX;
                real_max = -FLT_MAX;
                for(j = 0; j < (int)real_segments.size(); j++)
                    for(i = 0; i < (int)real_segments[j].size(); i++)
                    {
                        if(real_segments[j][i].x > real_max.x)    real_max.x = real_segments[j][i].x;
                        if(real_segments[j][i].x < real_min.x)    real_min.x = real_segments[j][i].x;
                        if(real_segments[j][i].y > real_max.y)    real_max.y = real_segments[j][i].y;
                        if(real_segments[j][i].y < real_min.y)    real_min.y = real_segments[j][i].y;
                        if(real_segments[j][i].z > real_max.z)    real_max.z = real_segments[j][i].z;
                        if(real_segments[j][i].z < real_min.z)    real_min.z = real_segments[j][i].z;
                    }
                if(ez_ilda_pointer->maintain_dxf_origin)
                {
                    if(fabs(real_max.x) > real_size)    real_size = fabs(real_max.x);
                    if(fabs(real_min.x) > real_size)    real_size = fabs(real_min.x);
                    if(fabs(real_max.y) > real_size)    real_size = fabs(real_max.y);
                    if(fabs(real_min.y) > real_size)    real_size = fabs(real_min.y);
                    if(fabs(real_max.z) > real_size)    real_size = fabs(real_max.z);
                    if(fabs(real_min.z) > real_size)    real_size = fabs(real_min.z);
                    real_size *= 2;
                }
                else // find the new center of the universe
                {
                    real_offset.x = (real_max.x - ((real_max.x - real_min.x) / 2));
                    real_offset.y = (real_max.y - ((real_max.y - real_min.y) / 2));
                    real_offset.z = (real_max.z - ((real_max.z - real_min.z) / 2));
    
                    if(fabs(real_max.x - real_min.x) > real_size)    real_size = fabs(real_max.x - real_min.x);
                    if(fabs(real_max.y - real_min.y) > real_size)    real_size = fabs(real_max.y - real_min.y);
                    if(fabs(real_max.z - real_min.z) > real_size)    real_size = fabs(real_max.z - real_min.z);
                }
                real_scale = 65000.0 / real_size;
            } // end if(auto_scale_dxf)
            //--------------------------------------------------------------------
            for(j = 0; j < (int)real_segments.size(); j++)
            {
                for(i = 0; i < (int)real_segments[j].size(); i++)
                {
                    point.x = (short)((real_segments[j][i].x - real_offset.x) * real_scale);
                    point.y = (short)((real_segments[j][i].y - real_offset.y) * real_scale);
                    point.z = (short)((real_segments[j][i].z - real_offset.z) * real_scale);
                    point.c = real_segments[j][i].c;
                    if(real_segments[j][i].is_blank())
                        point.blank();
                    else
                        point.unblank();
                    //point.set_unique();
                    frame.push_back(point);
                    point.clear();
                }
                frame.set_rgb_from_palette();
                push_back(frame);
                frame.clear();
            }
            renumber();
        }
        else
        {
            ez_ilda_pointer->process_exception = true;
            bmp_printf(   &BMP
                        , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                        , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                        , BMP.white
                        , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "[%s] has no dxf files"
                        , dir.c_str()
                      );
            bmp_to_SDL_Surface(&BMP, screen);
            error = EZ_ILDA_FILE_OPEN_FAILED;
        }
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_ifstream_dxf(ifstream& in, bool append)
{
    ez_ilda_header          header       ;
    ez_ilda_frame           frame        ;
    ez_palette              palette      ;
    //------------------------------------------------------------------------
    if(!append)
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //------------------------------------------------------------------------
    frame.format        = EZ_ILDA_3D_FRAME_DATA;
    frame.palette_index = EZ_DXF;
    //------------------------------------------------------------------------
    if(    (error = frame.from_ifstream_dxf(in, header))
        == EZ_ILDA_OK
      )
    {
        frame.quantity = (u_short)frame.number_of_points();
        frame.sync_rgb_and_palette();
        push_back(frame);
        ez_ilda_pointer->palette_index = EZ_DXF;
        renumber();
    }
    //------------------------------------------------------------------------
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_shp_file(string file)
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".shp"
      )
        file += ".shp";

    ifstream in((EZ_ILDA_DXF_SHARE + file).c_str(), ios::in);
    if(in.is_open())
    {
        frame_index = 0;
        file_name   = file;
        error       = EZ_ILDA_OK;

        error       = from_ifstream_shp(in);
        in.close();
    }
    else
    {
        ez_ilda_pointer->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "[any key] %s failed to open"
                    , file.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_ifstream_shp(ifstream& in)
{
    ez_ilda_header  header  ;
    ez_ilda_frame   frame   ;
    ez_palette      palette ;
    //------------------------------------------------------------------------
    clear();
    num_2D_frames = 0;
    num_3D_frames = 0;
    //------------------------------------------------------------------------
    frame.format        = EZ_ILDA_2D_FRAME_DATA;
    frame.palette_index = EZ_DXF;
    //------------------------------------------------------------------------
    while(    (error = frame.from_ifstream_shp(in, header))
           == EZ_ILDA_OK
      )
    {
        frame.quantity = (u_short)frame.number_of_points();
        frame.sync_rgb_and_palette();
        push_back(frame);
    }
    //------------------------------------------------------------------------
    ez_ilda_pointer->palette_index = EZ_DXF;
    renumber();
    return error;
}

//############################################################################
void ez_ilda_frame_set::flatten_z()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).flatten_z();
    return;
}

//############################################################################
void ez_ilda_frame_set::z_order_points(unsigned short span)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).z_order_points(span);
    return;
}

//############################################################################
void ez_ilda_frame_set::unblank_all_points()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).unblank_all_points();
    return;
}

//############################################################################
void ez_ilda_frame_set::minimize_blank_points()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).minimize_blank_points();
    return;
}

//############################################################################
void ez_ilda_frame_set::remove_dwell_points()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).remove_dwell_points();
    return;
}

//############################################################################
void ez_ilda_frame_set::minimize_vectors()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).minimize_vectors();
    return;
}

//############################################################################
void ez_ilda_frame_set::minimize_points()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).minimize_points();
    return;
}

//############################################################################
void ez_ilda_frame_set::accentuate_the_positive()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accentuate_the_positive();
    return;
}

//############################################################################
void ez_ilda_frame_set::eliminate_the_negative()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).eliminate_the_negative();
    return;
}

//############################################################################
void ez_ilda_frame_set::explode_current_frame()
{
    int                i;
    ez_ilda_frame_set  results,
                       exploded = current_frame().find_segments();

    for(i = 0; i < frame_index; i++)
        results += at(i);

    for(i = 0; i < exploded.number_of_frames(); i++)
        results += exploded[i];

    for(i = frame_index + 1; i < number_of_frames(); i++)
        results += at(i);

    *this = results;
    renumber();
    return;
}

//############################################################################
void ez_ilda_frame_set::collapse_selected_frames()
{
    ez_ilda_frame     collapsed;
    int               first_selected_frame_index = -1;

    for(int i = 0; i < number_of_frames(); i++)
    {
        if(at(i).is_selected)
        {
            if(first_selected_frame_index == -1)
                first_selected_frame_index = i;
            collapsed += at(i);
        }
    }
    if(first_selected_frame_index != -1)
    {
        frame(first_selected_frame_index) = collapsed;
        frame(first_selected_frame_index).is_selected = false;
        delete_selected_frames();
        renumber();
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::fracture_segments()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).fracture_segments();
    return;
}

//############################################################################
void ez_ilda_frame_set::bond_segments()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).bond_segments();
    return;
}

//############################################################################
void ez_ilda_frame_set::reorder_segments()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).reorder_segments();
    return;
}

//############################################################################
void ez_ilda_frame_set::add_dwell_points()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).add_dwell_points();
    return;
}

//############################################################################
void ez_ilda_frame_set::add_span_points()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).add_span_points();
    return;
}

//############################################################################
void ez_ilda_frame_set::delete_redundant_frames()
{
    ez_ilda_frame_set reduced;
    for(int i = 0; i < number_of_frames(); i++)
        if(at(i).is_unique)
            reduced.push_back(at(i));
    reduced.renumber();
    *this = reduced;
    return;
}

//############################################################################
void ez_ilda_frame_set::optimize_points()
{
    if(number_of_frames())
    {
        int i,
            dwell_point_count = 0,
            dwell_point_total = 0,
            blank_dwell       = ez_ilda_pointer->wave_sample_rate / 220;
        //--------------------------------------------------------------------
        for(i = 0; i < number_of_frames(); i++)
            at(i).optimize_points();
        //--------------------------------------------------------------------
        intro.clear();
        for(i = 0; i < blank_dwell; i++)
            intro.push_back(ez_ilda_point()); // origin
        //--------------------------------------------------------------------
        intro += ez_ilda_segment(intro[0], front()[0]); // from origin to first point in first frame
        //--------------------------------------------------------------------
        for(i = 0; i < blank_dwell; i++)
            intro.push_back(front()[0]);
        //--------------------------------------------------------------------
        for(i = 0; i < (number_of_frames() - 1); i++)
        {
            at(i).transition_to_next_frame.clear();
            dwell_point_total = dwell_points(delta_angle(at(i)[at(i).number_of_points() - 2] | at(i + 1).front(),
                                                         at(i)[at(i).number_of_points() - 2] | at(i).back(),
                                                         at(i).back()                        | at(i + 1).front()
                                                        )
                                            );
            if(dwell_point_total)
                for(dwell_point_count = 0; dwell_point_count < dwell_point_total; dwell_point_count++)
                    at(i).transition_to_next_frame.push_back(at(i).back());
            //----------------------------------------------------------------            
            at(i).transition_to_next_frame += ez_ilda_segment(at(i).back(), at(i + 1)[0]);
        }
        //--------------------------------------------------------------------
        back().transition_to_next_frame = ez_ilda_segment(back().back(), intro[0]);
        for(i = 0; i < blank_dwell; i++)
            back().transition_to_next_frame.push_back(intro[0]);
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::rainbow_recolor(int effect, float factor)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).rainbow_recolor(effect, factor);
    return;
}

//############################################################################
void ez_ilda_frame_set::quarter_turn(int plane, int turns)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).quarter_turn(plane, turns);
    return;
}

//############################################################################
int ez_ilda_frame_set::scale_around_origin(ez_3D_float factor)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).scale_around_origin(factor);
    return all_in_bounds;
}

//############################################################################
int ez_ilda_frame_set::scale_around_origin()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).scale_around_origin();
    return all_in_bounds;
}

//############################################################################
int ez_ilda_frame_set::rotate_around_origin(ez_3D_float angle)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).rotate_around_origin(angle);
    return all_in_bounds;
}

//############################################################################
int ez_ilda_frame_set::rotate_around_origin()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).rotate_around_origin();
    return all_in_bounds;
}

//############################################################################
int ez_ilda_frame_set::move(ez_3D_float offset)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).move(offset);
    return all_in_bounds;
}

//############################################################################
int ez_ilda_frame_set::move()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).move();
    return all_in_bounds;
}

//############################################################################
void ez_ilda_frame_set::toggle_frame_select_start()
{
    int i;     
    if(frame_select_start == -1) // was unset
    {
        frame_select_start = frame_index; // set
        if(frame_select_end != -1)
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = true;
            at(frame_select_end).is_selected = true;
        }
    }
    else // was set
    {
        if(frame_select_end != -1) // was set
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = false;
            at(frame_select_end).is_selected = false;
        }
        frame_select_start = -1; // unset
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::toggle_frame_select_end()
{
    int i;
    if(frame_select_end == -1) // was unset
    {
        frame_select_end = frame_index; // set
        if(frame_select_start != -1)
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = true;
            at(frame_select_end).is_selected = true;
        }
    }
    else // was set
    {
        if(frame_select_start != -1) // was set
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = false;
            at(frame_select_end).is_selected = false;
        }
        frame_select_end = -1; // unset
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::delete_selected_frames()
{
    ez_ilda_frame_set reduced;
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(!at(i).is_selected)
            reduced.push_back(at(i));
    }
    reduced.renumber();

    if(    frame_select_start != -1
        && frame_select_end   != -1
      )
    {
        frame_index = (frame_select_start - 1);
        frame_select_start = -1;
        frame_select_end   = -1;
    }

    if(frame_index < 0)
        frame_index = 0;

    if(frame_index > reduced.number_of_frames())
        frame_index = reduced.number_of_frames() - 1;

    *this = reduced;
    return;
}

//############################################################################
void ez_ilda_frame_set::trim_to_selected_frames()
{
    ez_ilda_frame_set reduced;
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(at(i).is_selected)
            reduced.push_back(at(i));
    }
    reduced.renumber();
    if(    frame_select_start != -1
        && frame_select_end   != -1
      )
    {
        frame_select_start = -1;
        frame_select_end   = -1;
    }
    frame_index = 0;
    *this = reduced;
    return;
}

//############################################################################
void ez_ilda_frame_set::bit_reduce_to_palette()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).bit_reduce_to_palette();
    return;
}

//############################################################################
void ez_ilda_frame_set::promote_to_true_color()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).promote_to_true_color();
    return;
}

//############################################################################
void ez_ilda_frame_set::best_match_palette(int index)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).best_match_palette(index);
    return;
}

//############################################################################
void ez_ilda_frame_set::best_match_target_palette()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).best_match_target_palette();
    return;
}

//############################################################################
void ez_ilda_frame_set::accept_palette(int index)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accept_palette(index);
    return;
}

//############################################################################
void ez_ilda_frame_set::accept_target_palette()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accept_target_palette();
    return;
}

//############################################################################
void ez_ilda_frame_set::color_over_ezfb_bitmap(struct ezfb_bitmap* bmp)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).color_over_ezfb_bitmap(bmp);
    return;
}

//############################################################################
void ez_ilda_frame_set::color_over_ezfb_bitmap(string file)
{
    char file_name[81];
    struct ezfb_bitmap bmp;
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".bmp"
      )
        file += ".bmp";
    strcpy(file_name, (EZ_ILDA_BMP_SHARE + file).c_str());
    if(bmp_from_file(&bmp, file_name))
    {
        color_over_ezfb_bitmap(&bmp);
        bmp_free(&bmp);
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::flip(int plane)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).flip(plane);
    return;
}

//############################################################################
void ez_ilda_frame_set::renumber()
{
    int   i,
          local_palette_index  = EZ_ILDA_DEFAULT,
          added_palettes       = 0,
          frame_id             = 0;
    bool  in_true_color        = false;
    char  number[7];
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
    {
        if((at(i).quantity = at(i).number_of_points()) > 1)
        {
            if(at(i).palette_index == EZ_TRUE_COLOR)
            {
                if(!in_true_color)
                {
                    frame_id++;
                    added_palettes++; // add EZFB_332
                    local_palette_index = EZ_EZFB_332;
                    in_true_color = true;
                }
                frame_id++;
                added_palettes++; // add true_color
            }
            else if(at(i).palette_index != local_palette_index)
            {
                frame_id++;
                added_palettes++;
                local_palette_index = at(i).palette_index;
                in_true_color = false;
            }
            sprintf(number, "%06d", frame_id);
            at(i).identity = frame_id;
            at(i).name     = "ES" + string(number);
            frame_id++;
        }
        else
            delete_frame(i);
    }
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
        at(i).total = (u_short)(number_of_frames() + added_palettes);
    if(frame_index >= number_of_frames())
        frame_index = number_of_frames() - 1;
    return;
}

//############################################################################
ez_ilda_frame_set& ez_ilda_frame_set::reverse()
{
    ez_ilda_frame_set reversed;
    for(int i = 1; i <= number_of_frames(); i++)
        reversed.push_back(at(number_of_frames() - i));
    clear();
    insert(begin(), reversed.begin(), reversed.end());
    renumber();
    return *this;
}

//############################################################################
ez_ilda_frame_set& ez_ilda_frame_set::frame_reverse()
{
    for(int i = 1; i <= number_of_frames(); i++)
        at(i).reverse();        
    return *this;
}

//############################################################################
ez_ilda_frame_set& ez_ilda_frame_set::superimpose_from_file(string file)
{
    ez_ilda_frame_set  overlay,
                       sum    ;

    int                i;
    overlay.from_file(file);

    if(overlay.number_of_frames() > 1 && number_of_frames() > 1)
    {
        if(number_of_frames() > overlay.number_of_frames())
        {
            for(i = 0; i < overlay.number_of_frames(); i++)
                sum += (frame(i) + overlay.frame(i));

            for(i = overlay.number_of_frames(); i < number_of_frames(); i++)
                sum += frame(i);
        }
        else if(number_of_frames() < overlay.number_of_frames())
        {
            for(i = 0; i < number_of_frames(); i++)
                sum += (frame(i) + overlay.frame(i));

            for(i = number_of_frames(); i < overlay.number_of_frames(); i++)
                sum += overlay.frame(i);
        }
        else
        {
            for(i = 0; i < number_of_frames(); i++)
                sum += (frame(i) + overlay.frame(i));
        }
    }
    else if(overlay.number_of_frames() == 1)
    {
        for(i = 0; i < number_of_frames(); i++)
            sum += (frame(i) + overlay.frame(0));
    }
    else if(number_of_frames() == 1)
    {
        for(i = 0; i < overlay.number_of_frames(); i++)
            sum += (frame(0) + overlay.frame(i));
    }
    *this = sum;
    return *this;
}

//############################################################################
void ez_ilda_frame_set::save_as_ilda(string file)
{
    ez_ilda_pointer->process_exception = false;
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".ild"
      )
        file += ".ild";

    ofstream out((EZ_ILDA_FRAME_SHARE + file).c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        to_ofstream(out);
        file_name = file;
        out.close();
    }
    else
    {
        ez_ilda_pointer->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "[any key] %s failed to open"
                    , file.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::to_ofstream(ofstream& out)
{
    int             file_section_number = 0,
                    local_palette_index = EZ_ILDA_DEFAULT;
    bool            in_true_color       = false;
    ez_ilda_header  header;
    //------------------------------------------------------------------------
    renumber();
    header.format  = EZ_ILDA_PALETTE_DATA;
    header.total   = front().total;    
    //------------------------------------------------------------------------    
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(at(i).palette_index == EZ_TRUE_COLOR)
        {
            if(!in_true_color)
            {
                header.name         = ez_ilda_pointer->palette_picker(EZ_EZFB_332).name;
                header.quantity     = ez_ilda_pointer->palette_picker(EZ_EZFB_332).number_of_colors();
                header.identity     = file_section_number;
                ez_ilda_pointer->palette_picker(EZ_EZFB_332).to_ofstream(out, header);
                local_palette_index = EZ_EZFB_332;
                file_section_number++;
                in_true_color = true;
            }
            file_section_number++;
        }
        else if(at(i).palette_index != local_palette_index)
        {
            header.name         = ez_ilda_pointer->palette_picker(at(i).palette_index).name;
            header.quantity     = ez_ilda_pointer->palette_picker(at(i).palette_index).number_of_colors();
            header.identity     = file_section_number;
            ez_ilda_pointer->palette_picker(at(i).palette_index).to_ofstream(out, header);
            local_palette_index = at(i).palette_index;
            file_section_number++;
            in_true_color = false;
        }
        at(i).is_2D(); // checks for z and sets format accordingly;
        at(i).to_ofstream(out);
        file_section_number++;
    }
    header.format   = EZ_ILDA_3D_FRAME_DATA ;
    header.name     = "ILDA_End"            ;
    header.quantity = 0                     ;
    header.identity = file_section_number   ;
    header.to_ofstream(out)                 ;
    return;
}

//############################################################################
void ez_ilda_frame_set::display_stats() const
{
    int y_line = 8 * ez_ilda_pointer->font_size_factor,
        x_off  = BMP.yres + EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor,
        y_off  = BMP.yres - 7 * y_line;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                , 0
                , ez_ilda_pointer->font_size_factor
                , "Frame set stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                , 0
                , ez_ilda_pointer->font_size_factor
                , "frames    %d"
                , number_of_frames()
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                , 0
                , ez_ilda_pointer->font_size_factor
                , "2D frames %d"
                , num_2D_frames
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                , 0
                , ez_ilda_pointer->font_size_factor
                , "3D frames %d"
                , num_3D_frames
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.make_rgb(ez_ilda_pointer->bg_color.r, ez_ilda_pointer->bg_color.g, ez_ilda_pointer->bg_color.b)
                , 0
                , ez_ilda_pointer->font_size_factor
                , "verticies %d"
                , total_points()
              );
    bmp_to_SDL_Surface(&BMP, screen);
    return;
}

//############################################################################
int ez_ilda_frame_set::total_wave_samples() const
{
    int   i,
          num_frame_scans,
          samples = intro.number_of_points();
          
    ez_ilda_segment bridge;          

    for(i = 0; i < number_of_frames(); i++)
    {
        if(at(i).number_of_points())
        {
            num_frame_scans = (int)ceil(   (   ez_ilda_pointer->wave_sample_rate 
                                             / at(i).number_of_points()
                                           )
                                         / ez_ilda_pointer->frames_per_second
                                       );
            bridge = ez_ilda_segment(at(i).last_lit_point(), at(i).first_lit_anchor());
            samples += (   (   (ez_ilda_pointer->frames_per_second > 0)
                             ? (num_frame_scans)
                             : (1)
                           ) * at(i).number_of_points()
                         + (num_frame_scans - 1) * bridge.number_of_points()
                         + at(i).transition_to_next_frame.number_of_points()
                       );
        }
    }
    return samples;
}

//############################################################################
void ez_ilda_frame_set::save_as_wave(string file) const
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".wav"
      )
        file += ".wav";
    ofstream out((EZ_ILDA_WAVE_SHARE + file).c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        if(number_of_frames() > 1)
            to_ofstream_wave(out);
        else
        {
            ez_ilda_frame copy(at(0));
            copy.rotate_around_origin();
            if(ez_ilda_pointer->auto_flatten_z)
                copy.flatten_z();
            if(ez_ilda_pointer->auto_optimize)
                copy.optimize_points();
            copy.save_as_wave(file);
        }
        out.close();
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::to_ofstream_wave(ofstream& out) const
{
    int i;     
    ez_ilda_frame_set copy(*this);
    for(i = 0; i < copy.number_of_frames(); i++)
        copy[i].rotate_around_origin();
    if(ez_ilda_pointer->auto_flatten_z)
        copy.flatten_z();
    if(ez_ilda_pointer->auto_optimize)
        copy.optimize_points();
    ez_wave_header  header(copy.total_wave_samples(), ez_ilda_pointer->wave_sample_rate);
    if(ez_ilda_pointer->auto_optimize)
    {
        header.ez_wave_mode = header.ez_wave_mode | EZ_WAVE_OPTIMIZED;
        header.parms.lit_dwell_overhang     = ez_ilda_pointer->lit_dwell_overhang;
        header.parms.distance_delta_max     = ez_ilda_pointer->distance_delta_max;
        header.parms.longest_dwell_microsec = ez_ilda_pointer->longest_dwell_microsec;
        header.parms.insignificant_distance = ez_ilda_pointer->insignificant_distance;
        header.parms.zero_angle_tolerance   = ez_ilda_pointer->zero_angle_tolerance;
        header.parms.frames_per_second      = ez_ilda_pointer->frames_per_second;        
    }
    header.to_ofstream(out);
    if(ez_ilda_pointer->auto_optimize)
        copy.intro.to_ofstream_wave(out);
    for(i = 0; i < copy.number_of_frames(); i++)
        copy.at(i).to_ofstream_wave(out);
    return;
}

//############################################################################
void ez_ilda_frame_set::save_as_bmp_directory(string dir) const
{
    char name[8];
#ifndef WIN32    
    mkdir((EZ_ILDA_BMP_SHARE + dir).c_str(), 0700);
#else    
    mkdir((EZ_ILDA_BMP_SHARE + dir).c_str());
#endif
    for(int i = 0; i < number_of_frames(); i++)
    {
        sprintf(name, "%06d", i);
        at(i).save_as_bmp(string(dir + '/' + name));
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::save_as_dxf_directory(string dir) const
{
    char name[8];
#ifndef WIN32    
    mkdir((EZ_ILDA_DXF_SHARE + dir).c_str(), 0700);
#else
    mkdir((EZ_ILDA_DXF_SHARE + dir).c_str());
#endif
    for(int i = 0; i < number_of_frames(); i++)
    {
        sprintf(name, "%06d", i);
        at(i).save_as_dxf(string(dir + '/' + name));
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::save_as_code(string object_name)
{
    ofstream out((EZ_ILDA_CPP_SHARE + object_name + ".cpp").c_str(), ios::out);
    if(out.is_open())
    {
        out << "//############################################################################\n"
               "// ~  This file is the product of LaserBoy!  ~\n"
               "// ~ International Laser Display Association ~\n"
               "// ~    ez_ilda file format driver in C++    ~\n"
               "//\n"
               "// by James Lehman\n"
               "// james@akrobiz.com\n"
               "//\n"
               "//############################################################################\n\n\n\n"
               "//############################################################################\n"
            << "ez_ilda_frame_set ";
        out << object_name;
        out << "()\n"
               "{\n"
               "    ez_ilda_frame     frame;\n"
               "    ez_ilda_frame_set frames;\n"
               "    int               frame_index;\n\n"
               "    for(frame_index = 0; frame_index < ";
        out << number_of_frames();
        out << "; frame_index++)\n"
               "    {\n";
        for(int i = 0; i < number_of_frames(); i++)
        {

            at(i).to_ofstream_code(out, "        ");            
            out << "        frames.push_back(frame);\n"
                   "        frame.clear();\n\n";
        }
        out << "    }\n"
               "    return frames;\n"
               "}\n\n"
               "//############################################################################\n"
               "//////////////////////////////////////////////////////////////////////////////\n"
               "//############################################################################\n\n\n";

        out.close();
    }
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
